// assemble with bass v18
architecture nes.cpu

output "Digital Devil Story - Megami Tensei (MMC3).nes", create

// Input ROM prepatched with the following translation:
// https://www.romhacking.net/translations/2287/
// 393232 bytes, CRC32 AC5ACC89, SHA-1 0C4FF5EB16A0826A92B6196EE67D8AEA2CB7572B
insert "Digital Devil Story - Megami Tensei.nes"

macro bank(index, address) {
  origin {index} * $2000 + $10
  base {address}
}

macro seek(address) {
  assert(pc() <= {address})
  ds {address} - pc()
}

// replace mapper type NAMCOT-3446 with MMC3
origin $06
  db $40, $00

// banks $1E-$1F will occupy the fixed region $C000-$FFFF
bank($1E, $C000)

seek($C000)
  // set vertical mirroring
  sta $A000
  nop

assert(pc() == $C004)

seek($C097)
  // update target for moved routine
  jsr sub_C67F

seek($C0EB)
  // update target for moved routine
  jsr sub_C67F

seek($C640)
  lda $0D
  beq +
  // replace inline code with call to common subroutine (saving space)
  jsr common
+
  lda $0F
  beq +
  sec
  sbc #$01
  asl
  tay
  // replace writes to R2-3 with writes to R0-R1
  ldx #$00
  stx $8000
  lda $C6BE,y
  sta $8001
  inx
  stx $8000
  lda $C6BF,y
  sta $8001
+
  rts

sub_C67F:
  lda $0E
  beq +
  lda $1A
  cmp #$06
  beq +
  ldx #$00
  ldy #$00
-
  iny
  bne -
  inx
  cpx #$02
  bne -
-
  bit $2002
  bvc -
  ldx $0136
-
  dex
  bne -
  lda $0E

common:
  sec
  sbc #$01
  asl
  tay
  // replace writes to R4-5 with writes to R2-R5
  ldx #$02
  stx $8000
  lda $C726,y
  sta $8001
  inx
  stx $8000
  ora #$01
  sta $8001
  inx
  stx $8000
  lda $C727,y
  sta $8001
  inx
  stx $8000
  ora #$01
  sta $8001
+
  rts

// shift CHR bank lookup table entries
seek($C6BE)
variable x = $B0
while x > 0 {
  db read(origin()) << 1
  x = x - 1
}

seek($FFF0)
reset:
  // first three instructions moved from $C000
  sei
  cld
  lda #$10
  // fix PRG ROM bank at $C000
  sta $8000
  jmp $C000

assert(pc() <= $FFFA)

seek($FFFC)
  // move reset handler to guaranteed fixed bank at $E000
  dw reset
